package trends;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.DataInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Queue;
import java.util.Set;

import twitter4j.HashtagEntity;
import twitter4j.IDs;
import twitter4j.Paging;
import twitter4j.RateLimitStatus;
import twitter4j.RateLimitStatusEvent;
import twitter4j.RateLimitStatusListener;
import twitter4j.ResponseList;
import twitter4j.Status;
import twitter4j.Twitter;
import twitter4j.TwitterException;
import twitter4j.TwitterFactory;
import twitter4j.User;
import twitter4j.internal.org.json.JSONArray;
import twitter4j.internal.org.json.JSONException;
import twitter4j.internal.org.json.JSONObject;

/**
 * 
 * @author karthik
 * 
 * @input 
 * 	1) filename containing the list of users with full path or current path
 *  2) number of tweets or retweets for each user
 *  
 * @output creates two files with twitterData.json with tweets and userData.txt with user information
 *  
 */
public class RetweetNetwork {
	
	static BufferedWriter outputTweet = null;
	static BufferedWriter outputUser = null;
	static int cascadeSize = 0;
	static int localCount = 0;
	
	private static void openFile() {
		File file = new File("twitterData.json");		
		File fileUser = new File("userData.json");
		try {
			outputTweet = new BufferedWriter(new FileWriter(file, true));
			outputUser = new BufferedWriter(new FileWriter(fileUser, true));
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	
	private static void storeJSON(Status status, boolean retweet, int size, int depth, int localCount) {	        
        try {       
        	JSONObject wrapper = new JSONObject();
        	wrapper.put("screen_name", status.getUser().getScreenName());
        	wrapper.put("text",status.getText().replace("\n", ""));
        	wrapper.put("retweet",retweet);
        	wrapper.put("retweet_count",status.getRetweetCount());
        	wrapper.put("created_at",status.getCreatedAt());
        	wrapper.put("cascade_size",size); 
        	wrapper.put("cascade_depth",depth); 
        	wrapper.put("local_count",localCount);
        	JSONArray list = new JSONArray();
        	HashtagEntity[] slist = status.getHashtagEntities();        	
        	for(HashtagEntity en : slist) {
        		list.put(en.getText());
        	}
        	
        	wrapper.put("hashtags", list);        	
        	System.out.println(wrapper.toString());
        	outputTweet.write(wrapper.toString() + "\n");        	 
        } catch (IOException e) {
                       
        } catch (JSONException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
    }
	
	private static void writeUserInfo(UserInfo user) {	        
        try {        	          	
        	JSONObject wrapper = user.getJSON();
          	outputUser.write(wrapper.toString() + "\n");  	
        } catch (IOException e) {
                       
        } 
    }
	
	private static void closeFile() {
		if (outputTweet != null) {
            try {
                outputTweet.close();
            } catch (IOException ignore) {
            }
        }
		
		if (outputUser != null) {
            try {
                outputUser.close();
            } catch (IOException ignore) {
            }
        }
	}

	private int calculateDepth(Twitter twitter, Set<Long> retweet_set, Set<Long> seed_follower) {
		int depth = 0;
		long cursor = -1;
		IDs ids = null;
		long water_mark = 0;
		long new_mark = 0;
		Set<Long> follower_list = new HashSet<Long>();
		Queue<Long> depthQ = new LinkedList<Long>();
		
		//intialize class variables
		localCount = 0;
		cascadeSize = 0;

		for(long id: seed_follower) {				
			if (retweet_set.contains(id)) {
				//remove it from the set
				retweet_set.remove(id);
				depthQ.add(id);
				new_mark = id;
				localCount++;
				cascadeSize++;
			}			
			
			if (retweet_set.isEmpty()) {
				depth++;				
				return depth;
			}
		}
		
		depth++;
						
		while(!depthQ.isEmpty()){	
			follower_list = new HashSet<Long>();
			cursor = -1;
			ids = null;
			long curr_id = depthQ.poll();
			//System.out.println("not an empty set" + curr_id);			
			do {
				try {				
					checkRateLimitStatus(twitter);
					ids = twitter.getFollowersIDs(curr_id, cursor);	
					
					for(long id: ids.getIDs()){
						follower_list.add(new Long(id));
					}
				} catch (TwitterException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();				
				}				
			} while ((cursor = ids.getNextCursor()) != 0);			
			
			for(long id: follower_list) {				
				if (retweet_set.contains(id)) {
					//remove it from the set
					retweet_set.remove(id);
					depthQ.add(id);
					new_mark = id;
					cascadeSize++;
				}			
				
				if (retweet_set.isEmpty()) {
					depth++;
					return depth;
				}
			}
			
			if (curr_id == water_mark){
				depth++;
				water_mark = new_mark;
			}
		}
		return depth;
	}
	
	private int getLocalInfluence(Set<Long> ids, Set<Long> retweet_set) {
		int local_count = 0;
				
		for(long id: ids) {		
			//System.out.println("follower_id" + id);
			if (retweet_set.contains(id)) {
				//remove it from the set
				//retweet_set.remove(id);						
				local_count++;
			}	
			
		}
					
		return local_count;
	}
	
	public Set<Long> updateCascade (Twitter twitter, long tweet_id) {
		int cascadeSize = 0;
		int page = 1;
		Set<Long> retweet_set = new HashSet();
		// get the user who retweets the user 
		try {		
			ResponseList<User> retweet_users;
			
			do {
				checkRateLimitStatus(twitter);
				retweet_users = twitter.getRetweetedBy(tweet_id, new Paging(page, 100));
				cascadeSize = retweet_users.size();				
                for (User user : retweet_users) {                	
                	retweet_set.add(user.getId());	                    
                }
                page++;
            } while (retweet_users.size() != 0);
	
		} catch (TwitterException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}		
		
		return retweet_set;
	}
	
	public void dataExtractor (Twitter twitter,String userName,int count) {				
		boolean userStored = true;
		Paging page = new Paging(1);
		
        page.setCount(count);
    	UserInfo user = new UserInfo(userName,0,0,0);    	
    	Set<Long> follower_list = new HashSet<Long>();
    	
    	System.out.println("Showing tweets by @" + userName + "");
         
    	List<Status> statuses = null;
		try {			
			IDs ids = null;
			long cursor = -1;
			
			do {
				try {
					checkRateLimitStatus(twitter);
					ids = twitter.getFollowersIDs(userName, cursor);										
					for(long id: ids.getIDs()){
						follower_list.add(new Long(id));
					}
				} catch (TwitterException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}				
			} while ((cursor = ids.getNextCursor()) != 0);			
			
			checkRateLimitStatus(twitter);
			statuses = twitter.getUserTimeline(userName, page);
		         					
			checkRateLimitStatus(twitter);
			for (Status status : statuses) {				
	             //System.out.println("@" + status.getUser().getScreenName() + " - " + status.getText());
	         		if (userStored) {	         			
	         			user = new UserInfo(status.getUser().getScreenName(),
     								status.getUser().getFollowersCount(),status.getUser().getFriendsCount(),status.getUser().getListedCount());	         						
	         			userStored = false;
	         		}
	         		
	         		Set<Long> retweet_set = new HashSet();
	         		retweet_set = updateCascade(twitter,status.getId());
	         		//int depth = calculateDepth(twitter,retweet_set,follower_list);
	         		int depth = 0;
	         		localCount = getLocalInfluence(follower_list,retweet_set);
	         		//int depth = getLocalInfluence(follower_list,retweet_set);
	         		storeJSON(status, false, cascadeSize, depth, localCount);					
	         }
	         
	         page = new Paging(1);
	         page.setCount(count);
	         checkRateLimitStatus(twitter);
	         statuses = twitter.getRetweetedByUser(userName, page);
	         System.out.println("Showing retweets by @" + userName + "");
	         for (Status status : statuses) {	        	 
	             //System.out.println("@" + status.getUser().getScreenName() + " - " + status.getText());
	             storeJSON(status, true, 0, 0, 0);	             	                          
	         }
	         
	         writeUserInfo(user);
		} catch (TwitterException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	
	private void checkRateLimitStatus(Twitter twitter) {
		RateLimitStatus status;
		try {
			status = twitter.getRateLimitStatus();
			 if (status.getRemainingHits() < 1) {
		            try {                
		            	System.out.println("Limit reached going to sleep");
		                Thread.currentThread().sleep(status.getSecondsUntilReset() * 1000);               
		            } catch (InterruptedException ie) {
		                ie.printStackTrace();
		            }
		        }
		} catch (TwitterException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}       
    }
	
	public static void main(String[] args) {
        if (args.length < 2) {
            System.out.println("Usage: user file and number");
            System.exit(-1);
        }
        
        Twitter twitter = new TwitterFactory().getInstance();                        
        openFile();
        int count = Integer.parseInt(args[1]);
        RetweetNetwork net = new RetweetNetwork();
        
        try {
	        FileInputStream fstream = new FileInputStream(args[0]);
	        DataInputStream in = new DataInputStream(fstream);
	        BufferedReader br = new BufferedReader(new InputStreamReader(in));
	        String strLine = null;
	        //Read File Line By Line
	        while ((strLine = br.readLine()) != null)   {
	            // Print the content on the console
	        	//System.out.println(strLine);
	        	net.dataExtractor(twitter, strLine, count);	      	            		       
	        }
	        
	        //Close the input stream
	        in.close();   
	        closeFile();
	        System.out.println("done.");
        }catch (Exception e){//Catch exception if 
        	  System.err.println("Error: " + e.getMessage());
        }
        System.exit(0);
    }
}
